import os
import threading
from abc import abstractmethod, ABCMeta
from typing import List

from sidecar.cloud_logger.logs import ILogEntry


LogPath = "/var/ftp/sandbox/logs"


class ICloudLogger(metaclass=ABCMeta):

    @classmethod
    @abstractmethod
    def write(cls, log_entry: ILogEntry):
        pass


class FileLogger(ICloudLogger):
    def __init__(self):
        self._lock = threading.RLock()

    def write(self, log_entry: ILogEntry):
        with self._lock:
            if not os.path.exists(LogPath):
                os.makedirs(LogPath)

            with open(log_entry.get_log_filepath(LogPath), "ab") as stream:
                for time, line in log_entry.log_events:
                    line_bytes = "[{LOG_TYPE}][{TIME}]{LINE}\n" \
                        .format(LOG_TYPE=log_entry.log_type, TIME=time.strftime('%Y-%m-%d %H:%M:%S'), LINE=line) \
                        .encode('utf8')
                    stream.write(line_bytes)


class FakeFileLogger(ICloudLogger):

    def __init__(self):
        super().__init__()
        self.entries = []

    def write(self, log_entry: ILogEntry):
        self.entries.append(log_entry)

    def get_entries(self) -> List[ILogEntry]:
        return self.entries



from enum import Enum
from typing import List

from sidecar.model.objects import ISidecarService
from sidecar.services.service_status_state import ServiceStatusState, ServiceStatus


class CanStartTerminationProcessResult:
    TRUE = 'true'
    WAIT = 'wait'
    DEPENDENCIES_FAILURE = 'dependencies_failure'


class ServiceTerminationStartPolicy:
    def __init__(self, services: List[ISidecarService], service_status_state: ServiceStatusState):
        self._service_status_state = service_status_state

        service_names = set(s.name for s in services)
        # in teardown only dependencies between services are supported
        self._service_dependencies = {service.name: list([d for d in service.dependencies if d in service_names]) for
                                      service in services}

    def can_start_termination_process(self, service_name: str) -> str:
        dependencies = self._service_dependencies.get(service_name, [])
        service_statuses = self._service_status_state.get_status(service_names=dependencies)

        # ignore all services with validation failed status
        dependencies_status = [status for status in service_statuses.values()
                               if status != ServiceStatus.VALIDATION_FAILED]

        if not dependencies_status:
            return CanStartTerminationProcessResult.TRUE

        if all(status == ServiceStatus.TERMINATED for status in dependencies_status):
            return CanStartTerminationProcessResult.TRUE

        if any(status == ServiceStatus.TERMINATE_FAILED for status in dependencies_status):
            return CanStartTerminationProcessResult.DEPENDENCIES_FAILURE

        return CanStartTerminationProcessResult.WAIT



import threading
from logging import Logger
from typing import List, Dict

from blinker import signal

from sidecar.const import AppNetworkStatus, Signals
from sidecar.utils import CallsLogger


class AppHealthCheckState:
    def __init__(self, app_names: List[str], logger: Logger):
        self._lock = threading.RLock()
        self._logger = logger
        self._app_states = {app_name: AppNetworkStatus.PENDING for app_name in app_names}  # type: Dict[str, str]

    @CallsLogger.wrap
    def set_app_state(self, app_name: str, status: str):
        with self._lock:
            self._app_states[app_name] = status

            all_apps_complete = all(AppNetworkStatus.is_end_status(status)
                                    for status
                                    in self._app_states.values())

        if all_apps_complete:
            signal(Signals.ON_APPS_DEPLOYMENT_COMPLETE).send(self)

    @CallsLogger.wrap
    def all_complete_with_success(self) -> bool:
        with self._lock:
            return all(status == AppNetworkStatus.COMPLETED
                       for status
                       in self._app_states.values())

    @CallsLogger.wrap
    def get_apps_state(self, app_names: List[str]) -> Dict[str, str]:
        with self._lock:
            return {k: v for k, v in self._app_states.items() if k in app_names}



from abc import ABCMeta, abstractmethod
from logging import Logger
from typing import List

from sidecar.app_instance_identifier import AppInstanceIdentifier
from sidecar.sandbox_error import SandboxError


class StaleAppInstanceException(Exception):
    pass


class IAppInstanceService:
    __metaclass__ = ABCMeta

    def __init__(self, logger: Logger):
        self._logger = logger

    @abstractmethod
    def update_status_if_not_stale(self, app_instance_identifier: AppInstanceIdentifier, status: str):
        raise NotImplementedError

    @abstractmethod
    def update_artifacts_status(self, app_instance_identifier: AppInstanceIdentifier, status: str):
        raise NotImplementedError

    @abstractmethod
    def update_configuration_status(self, app_instance_identifier: AppInstanceIdentifier, status: str):
        raise NotImplementedError

    @abstractmethod
    def check_which_exist(self, identifiers: List[AppInstanceIdentifier]) -> List[AppInstanceIdentifier]:
        raise NotImplementedError

    @abstractmethod
    def get_public_address(self, app_instance_identifier: AppInstanceIdentifier) -> str:
        raise NotImplementedError

    @abstractmethod
    def update_deployment_outputs(self, app_instance_identifier: AppInstanceIdentifier, outputs: {}):
        raise NotImplementedError

    @abstractmethod
    def get_deployment_outputs(self, app_name: str) -> {}:
        raise NotImplementedError

    @abstractmethod
    def is_qualiy_off(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def add_error(self, app_instance_identifier: AppInstanceIdentifier, error: SandboxError):
        raise NotImplementedError



import os
import pathlib

from sidecar.cloud_logger.file_logger import LogPath


class ServiceFileStore:

    @staticmethod
    def save_execution_output(name: str, cmd: str, output: bytes):
        if not os.path.exists(LogPath):
            pathlib.Path(LogPath).mkdir(parents=True, exist_ok=True)
        with open(os.path.join(LogPath, f"{name}.{cmd}.log"), "wb") as f:
            f.write(output)



import datetime
import threading
from logging import Logger
from timeit import default_timer as timer
from time import sleep

from retrying import retry

from sidecar.app_instance_service import IAppInstanceService
from sidecar.status_maintainer import StatusMaintainer


class QualiyMonitor:
    SAMPLE_INTERVAL = 10

    def __init__(self,
                 status_maintainer: StatusMaintainer,
                 app_instance_service: IAppInstanceService,
                 logger: Logger):
        self._status_maintainer = status_maintainer
        self._app_instance_service = app_instance_service
        self._logger = logger
        self._lock = threading.RLock()
        self._thread = None

    def start(self):
        with self._lock:
            if not self._thread or not self._thread.is_alive():
                self._thread = threading.Thread(target=self.sampling_loop, daemon=True)
                self._thread.start()
            else:
                self._logger.warning('Sampling loop is already running')

    def sampling_loop(self):
        self._logger.info(f'Waiting for qualiy to be turned off ({self.SAMPLE_INTERVAL} sec intervals)')
        start = timer()
        iterations = 0
        while not self.is_qualiy_off():
            iterations += 1
            sleep(self.SAMPLE_INTERVAL)

        duration = datetime.timedelta(seconds=timer()-start)
        self._logger.info(f'Qualiy is off (took {duration}, and {iterations} valid iterations)')
        try:
            self._status_maintainer.update_qualiy_status('off')
        except:
            self._logger.exception('Failed to set qualiy status "off"')

    @retry(wait_exponential_multiplier=1000, wait_exponential_max=1000*60*30)  # Exponential Backoff, starts from 1 sec to 30 min
    def is_qualiy_off(self) -> bool:
        try:
            return self._app_instance_service.is_qualiy_off()
        except:
            self._logger.warning('Failed to get qualiy instance power status', exc_info=True)
            raise



import json
import threading
from abc import ABCMeta, abstractmethod
from logging import Logger

from sidecar.aws_status_maintainer import AWSStatusMaintainer
from sidecar.azure_clp.azure_status_maintainer import AzureStatusMaintainer
from sidecar.const import Const
from sidecar.kub_api_pod_reader import KubApiPodReader
from sidecar.kub_api_service import IKubApiService
from sidecar.services.service_status_state import ServiceStatusState
from sidecar.utils import Utils, CallsLogger


class IServiceUpdater(metaclass=ABCMeta):
    def __init__(self, service_status_state: ServiceStatusState, logger: Logger):
        self.service_status_state = service_status_state
        self.logger = logger
        self.lock = threading.RLock()

    @CallsLogger.wrap
    def update_status(self, name: str, status: str):
        self.on_update_status(name=name, status=status)
        self.service_status_state.update_status(name=name, status=status)

    @abstractmethod
    def on_update_status(self, name: str, status: str):
        raise NotImplemented()

    @abstractmethod
    def update_deployment_outputs(self, service_name: str, outputs: {}):
        raise NotImplemented()

    @abstractmethod
    def get_deployment_outputs(self, service_name: str) -> {}:
        raise NotImplemented()


class K8SServiceUpdater(IServiceUpdater):

    def __init__(self, kub_api_service: IKubApiService, service_status_state: ServiceStatusState, logger: Logger):
        super().__init__(service_status_state, logger)
        self._k8s_service = kub_api_service

    @staticmethod
    def _can_update_pod(pod):
        return pod and not KubApiPodReader.is_pod_ended(pod) and not KubApiPodReader.is_pod_terminating(pod)

    @CallsLogger.wrap
    def on_update_status(self, name: str, status: str):
        with self.lock:
            pod = self._get_pod()
            service_data = KubApiPodReader.get_service_json(pod, name)
            service_data["status"] = status
            annotation_change = {name: json.dumps(service_data)}
            self._k8s_service.update_pod(Const.SERVICE_EXECUTION_POD_NAME, annotation_change)

    def _get_pod(self):
        pod = Utils.retry_on_exception(
            func=lambda: self._k8s_service.get_pod_by_name(name=Const.SERVICE_EXECUTION_POD_NAME),
            logger=self.logger,
            logger_msg="trying to get service execution pod")
        return pod

    def update_deployment_outputs(self, service_name: str, outputs: {}):
        with self.lock:
            pod = self._get_pod()
            service_data = KubApiPodReader.get_service_json(pod, service_name)
            service_data["outputs"] = outputs
            annotation_change = {service_name: json.dumps(service_data)}
            self._k8s_service.update_pod(Const.SERVICE_EXECUTION_POD_NAME, annotation_change)

    def get_deployment_outputs(self, service_name: str) -> {}:
        pod = self._get_pod()
        if service_name in KubApiPodReader.get_pod_annotations(pod):
            service = KubApiPodReader.get_service_json(pod, service_name)
            return service.get('outputs', {})

        raise Exception(f"could not find service with name '{service_name}'")


class AwsServiceUpdater(IServiceUpdater):

    def __init__(self, status_maintainer: AWSStatusMaintainer, service_status_state: ServiceStatusState,
                 logger: Logger):
        super().__init__(service_status_state, logger)
        self._status_maintainer = status_maintainer

    @CallsLogger.wrap
    def on_update_status(self, name: str, status: str):
        with self.lock:
            self._status_maintainer.update_service_status(name=name, status=status)

    def update_deployment_outputs(self, service_name: str, outputs: {}):
        with self.lock:
            return self._status_maintainer.update_service_outputs(service_name, outputs)

    def get_deployment_outputs(self, service_name: str) -> {}:
        return self._status_maintainer.get_service_deployment_outputs(service_name)


class AzureServiceUpdater(IServiceUpdater):

    def update_deployment_outputs(self, service_name: str, outputs: {}):
        with self.lock:
            self._status_maintainer.update_service_outputs(name=service_name, outputs=outputs)

    def get_deployment_outputs(self, service_name: str) -> {}:
        return self._status_maintainer.get_service_deployment_outputs(service_name)

    def __init__(self, status_maintainer: AzureStatusMaintainer, service_status_state: ServiceStatusState,
                 logger: Logger):
        super().__init__(service_status_state, logger)
        self._status_maintainer = status_maintainer

    @CallsLogger.wrap
    def on_update_status(self, name: str, status: str):
        with self.lock:
            self._status_maintainer.update_service_status(name=name, status=status)
